@@ -7,6 +7,7 @@
 #include "computation/expression/expression.H"
 #include "computation/expression/let.H"
 #include "computation/expression/trim.H"
+#include "computation/operations.H"
 
 using std::string;
 using std::vector;
@@ -14,8 +15,7 @@ using std::map;
 using std::set;
 using std::pair;
 
-#include "computation/operations.H"
-#include "computation/computation.H"
+#include "args.H"
 
 long total_reductions = 0;
 long total_changeable_reductions = 0;
@@ -50,7 +50,7 @@ void throw_reg_exception(reg_heap& M, int t, int R, myexception& e)
 {
     dot_graph_for_token(M, t);
     string SSS = unlet(untranslate_vars(
-			   untranslate_vars(deindexify(trim_unnormalize(M.access(R).C)), M.get_identifiers()),
+			   untranslate_vars(deindexify(trim_unnormalize(M[R])), M.get_identifiers()),
 			   get_constants(M,t)
 			   )
 	).print();
@@ -65,6 +65,8 @@ class RegOperationArgs: public OperationArgs
 {
     const int S;
 
+    const int P;
+
     const closure& current_closure() const {return memory().closure_stack.back();}
 
     bool evaluate_changeables() const {return true;}
@@ -86,7 +88,10 @@ class RegOperationArgs: public OperationArgs
 	    // Note that although R2 is newly used, R3 might be already used if it was 
 	    // found from R2 through a non-changeable reg_var chain.
 	    if (M.reg_is_changeable(R3))
+	    {
+		used_changeable = true;
 		M.set_used_input(S, R3);
+	    }
 
 	    return value;
 	}
@@ -94,32 +99,36 @@ class RegOperationArgs: public OperationArgs
     const closure& evaluate_reg_to_closure(int R2)
 	{
 	    int R3 = evaluate_reg_to_reg(R2);
-	    return M.access(R3).C;
+	    return M[R3];
 	}
   
     const closure& evaluate_reg_to_closure_(int R2)
 	{
 	    int R3 = evaluate_reg_no_record(R2);
-	    return M.access(R3).C;
+	    return M[R3];
 	}
 
 public:
 
+    bool used_changeable = false;
+
     // If we unreference regs that evaluate to a variable, then we unreference p->let q=2 in q
     // and point references to q instead of p.  But then it would not be true that a variable can
     // only be referenced if the slot that created it is still referenced.
     
     int allocate_reg()
 	{
+	    int s = used_changeable?S:P;
 	    int r = OperationArgs::allocate_reg();
-	    M.mark_reg_created_by_step(r,S);
+	    if (s>0)
+		M.mark_reg_created_by_step(r,s);
 	    return r;
 	}
   
     RegOperationArgs* clone() const {return new RegOperationArgs(*this);}
 
-    RegOperationArgs(int s, reg_heap& m)
-	:OperationArgs(m), S(s)
+    RegOperationArgs(int s, int p, reg_heap& m)
+	:OperationArgs(m), S(s), P(p)
 	{ }
 };
 
@@ -207,8 +216,8 @@ int reg_heap::incremental_evaluate_unchangeable(int R)
 pair<int,int> reg_heap::incremental_evaluate_(int R)
 {
     assert(is_completely_dirty(root_token));
-    assert(is_valid_address(R));
-    assert(is_used(R));
+    assert(regs.is_valid_address(R));
+    assert(regs.is_used(R));
 
 #ifndef NDEBUG
     if (reg_has_value(R))
@@ -218,19 +227,19 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	assert(not E.head().is_a<expression>());
 	assert(not E.is_index_var());
     }
-    if (access(R).C.exp.is_index_var())
+    if (regs.access(R).C.exp.is_index_var())
 	assert(not reg_has_value(R));
 #endif
 
     while (1)
     {
-	assert(access(R).C.exp);
+	assert(regs.access(R).C.exp);
 
 #ifndef NDEBUG
-	//    std::cerr<<"   statement: "<<R<<":   "<<access(R).E.print()<<std::endl;
+	//    std::cerr<<"   statement: "<<R<<":   "<<regs.access(R).E.print()<<std::endl;
 #endif
 
-	reg::type_t reg_type = access(R).type;
+	reg::type_t reg_type = regs.access(R).type;
 
 	if (reg_type == reg::type_t::constant) return {R,R};
 
@@ -275,15 +284,15 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	}
 	else if (reg_type == reg::type_t::index_var)
 	{
-	    int index = access(R).C.exp.as_index_var();
-	    int R2 = access(R).C.lookup_in_env( index );
+	    int index = regs.access(R).C.exp.as_index_var();
+	    int R2 = regs.access(R).C.lookup_in_env( index );
 	    return incremental_evaluate(R2);
 	}
 	else
 	    assert(reg_type == reg::type_t::unknown);
 
 	/*---------- Below here, there is no call, and no value. ------------*/
-	if (access(R).C.exp.head().is_index_var())
+	if (regs.access(R).C.exp.head().is_index_var())
 	{
 	    assert( not reg_is_changeable(R) );
 
@@ -291,7 +300,7 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 
 	    assert( not reg_has_call(R) );
 
-	    access(R).type = reg::type_t::index_var;
+	    regs.access(R).type = reg::type_t::index_var;
 
 	    clear_result(R);
 	    int s = step_index_for_reg(R);
@@ -299,9 +308,9 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 		clear_back_edges_for_step(s);
 	    clear_step(R);
 
-	    int index = access(R).C.exp.as_index_var();
+	    int index = regs.access(R).C.exp.as_index_var();
 
-	    int R2 = access(R).C.lookup_in_env( index );
+	    int R2 = regs.access(R).C.lookup_in_env( index );
 
 	    // Return the end of the index_var chain.
 	    // We used to update the index_var to point to the end of the chain.
@@ -310,9 +319,9 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	}
 
 	// Check for WHNF *OR* heap variables
-	else if (is_WHNF(access(R).C.exp))
+	else if (is_WHNF(regs.access(R).C.exp))
 	{
-	    access(R).type = reg::type_t::constant;
+	    regs.access(R).type = reg::type_t::constant;
 	    clear_result(R);
 	    int s = step_index_for_reg(R);
 	    if (s > 0)
@@ -322,9 +331,9 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	}
 
 #ifndef NDEBUG
-	else if (access(R).C.exp.head().is_a<Trim>())
+	else if (regs.access(R).C.exp.head().is_a<Trim>())
 	    std::abort();
-	else if (access(R).C.exp.type() == parameter_type)
+	else if (regs.access(R).C.exp.type() == parameter_type)
 	    std::abort();
 #endif
 
@@ -336,12 +345,14 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	    if (not has_step(R))
 		add_shared_step(R);
 	    int S = step_index_for_reg(R);
+	    // FIXME - check that this agrees with our caller!
+	    int P = regs.access(R).created_by.first;
 
 	    try
 	    {
-		closure_stack.push_back (access(R).C );
-		RegOperationArgs Args(S, *this);
-		auto O = access(R).C.exp.head().assert_is_a<Operation>()->op;
+		closure_stack.push_back (regs.access(R).C );
+		RegOperationArgs Args(S, P, *this);
+		auto O = regs.access(R).C.exp.head().assert_is_a<Operation>()->op;
 		closure value = (*O)(Args);
 		closure_stack.pop_back();
 		total_reductions++;
@@ -349,12 +360,13 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 		    total_changeable_reductions++;
 
 		// If the reduction doesn't depend on modifiable, then replace E with the value.
-		if (steps[S].used_inputs.empty())
+		if (not Args.used_changeable)
 		{
 		    // The old used_input slots are not invalid, which is OK since none of them are changeable.
 		    assert(not reg_has_call(R) );
 		    assert(not reg_has_value(R));
 		    assert(step_for_reg(R).used_inputs.empty());
+		    assert(step_for_reg(R).created_regs.empty());
 		    set_C(R, std::move(value) );
 		}
 		else
@@ -371,11 +383,9 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 		    }
 		    else
 		    {
-			r2 = allocate();
+			r2 = Args.allocate( std::move(closure_stack.back()) ) ;
+			assert(regs.access(r2).created_by.first == S);
 			assert(not has_step(r2));
-			mark_reg_created_by_step(r2,S);
-			total_reg_allocations++;
-			set_C(r2, std::move(closure_stack.back()));
 		    }
 
 		    auto p = incremental_evaluate(r2);
@@ -391,12 +401,10 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 		    }
 		    else
 		    {
-			int r2 = allocate();
+			int r2 = Args.allocate( std::move(closure_stack.back()) );
 			assert(not has_step(r2));
-			mark_reg_created_by_step(r2,S);
-			total_reg_allocations++;
-			set_C(r2, std::move(closure_stack.back()));
-			access(r2).type = reg::type_t::constant;
+			regs.access(r2).type = reg::type_t::constant;
+			// assert(is_WHNF(regs.access(r2).C.exp)) ?
 			p = {r2,r2};
 		    }
 #endif
@@ -412,7 +420,7 @@ pair<int,int> reg_heap::incremental_evaluate_(int R)
 	    }
 	    catch (error_exception& e)
 	    {
-		throw e;
+		throw;
 	    }
 	    catch (myexception& e)
 	    {
@@ -438,6 +446,13 @@ void reg_heap::incremental_evaluate_from_call(int S, closure& value)
     incremental_evaluate_from_call_(S);
 }
 
+// The goal here is to merge steps S->S2->S3 => S.  At this point S is known to be changeable.
+//
+// However, S->S2 and S has performed allocations, then we don't want S2 to unshare those
+// if it is unshared.  Therefore once allocations have been performed we don't want to merge
+// with any downstream regs that are changeable.
+//
+// Therefore, if we have any allocations performed in S, create a new reg and delegate to it.
 void reg_heap::incremental_evaluate_from_call_(int S)
 {
     assert(is_completely_dirty(root_token));
@@ -445,7 +460,6 @@ void reg_heap::incremental_evaluate_from_call_(int S)
     assert(S > 0);
 
     while (not closure_stack.back().exp.head().is_index_var() and
-	   closure_stack.back().exp.head().type() != let2_type and
 	   not is_WHNF(closure_stack.back().exp))
     {
 #ifndef NDEBUG
@@ -453,22 +467,28 @@ void reg_heap::incremental_evaluate_from_call_(int S)
 	assert(closure_stack.back().exp.type() != parameter_type);
 #endif
 
+	// If any allocations have been performed by any of the steps merged into S
+	// then don't merge any children into S.
+	if (not steps[S].created_regs.empty())
+	{
+	    int r2 = allocate_reg_from_step(S,std::move(closure_stack.back()));
+	    closure_stack.back() = closure(index_var(0),{r2});
+	    return;
+	}
+
 	try
 	{
-	    RegOperationArgs Args(S, *this);
-	    int n_used_inputs1 = steps[S].used_inputs.size();
+	    RegOperationArgs Args(S, S, *this);
 	    auto O = closure_stack.back().exp.head().assert_is_a<Operation>()->op;
 	    closure_stack.back() = (*O)(Args);
-	    int n_used_inputs2 = steps[S].used_inputs.size();
-	    bool changed = n_used_inputs2 > n_used_inputs1;
 	
 	    total_reductions++;
-	    if (changed)
+	    if (Args.used_changeable)
 		total_changeable_reductions++;
 	}
 	catch (error_exception& e)
 	{
-	    throw e;
+	    throw;
 	}
 	catch (myexception& e)
 	{
@@ -481,15 +501,6 @@ void reg_heap::incremental_evaluate_from_call_(int S)
 	    throw_reg_exception(*this, root_token, closure_stack.back(), e);
 	}
     }
-    if (closure_stack.back().exp.head().type() == let2_type)
-    {
-	int r2 = allocate();
-	assert(not has_step(r2));
-	mark_reg_created_by_step(r2,S);
-	total_reg_allocations++;
-	set_C(r2, std::move(closure_stack.back()));
-	closure_stack.back() = closure(index_var(0),{r2});
-    }
 }
 
 /// These are LAZY operation args! They don't evaluate arguments until they are evaluated by the operation (and then only once).
@@ -497,7 +508,7 @@ class RegOperationArgsUnchangeable: public OperationArgs
 {
     const int R;
 
-    const closure& current_closure() const {return memory()[R].C;}
+    const closure& current_closure() const {return memory()[R];}
 
     bool evaluate_changeables() const {return false;}
 
@@ -517,10 +528,10 @@ class RegOperationArgsUnchangeable: public OperationArgs
     const closure& evaluate_reg_to_closure(int R2)
 	{
 	    int R3 = evaluate_reg_to_reg(R2);
-	    if (M.access(R3).type == reg::type_t::changeable)
+	    if (M.reg_type(R3) == reg::type_t::changeable)
 		throw no_context();
-	    assert(M.access(R3).type == reg::type_t::constant);
-	    return M.access(R3).C;
+	    assert(M.reg_type(R3) == reg::type_t::constant);
+	    return M[R3];
 	}
 
     const closure& evaluate_reg_to_closure_(int R2)
@@ -539,40 +550,40 @@ public:
 
 int reg_heap::incremental_evaluate_unchangeable_(int R)
 {
-    assert(is_valid_address(R));
-    assert(is_used(R));
+    assert(regs.is_valid_address(R));
+    assert(regs.is_used(R));
 
 #ifndef NDEBUG
-    assert(not access(R).C.exp.head().is_a<expression>());
+    assert(not regs.access(R).C.exp.head().is_a<expression>());
 #endif
 
     while (1)
     {
-	assert(access(R).C.exp);
+	assert(regs.access(R).C.exp);
 
-	reg::type_t reg_type = access(R).type;
+	reg::type_t reg_type = regs.access(R).type;
 
 	if (reg_type == reg::type_t::constant or reg_type == reg::type_t::changeable)
 	    break;
 
 	else if (reg_type == reg::type_t::index_var)
 	{
-	    int index = access(R).C.exp.as_index_var();
-	    int R2 = access(R).C.lookup_in_env( index );
+	    int index = regs.access(R).C.exp.as_index_var();
+	    int R2 = regs.access(R).C.lookup_in_env( index );
 	    return incremental_evaluate_unchangeable(R2);
 	}
 	else
 	    assert(reg_type == reg::type_t::unknown);
 
 	/*---------- Below here, there is no call, and no value. ------------*/
-	const int type = access(R).C.exp.head().type();
+	const int type = regs.access(R).C.exp.head().type();
 	if (type == index_var_type)
 	{
-	    access(R).type = reg::type_t::index_var;
+	    regs.access(R).type = reg::type_t::index_var;
 
-	    int index = access(R).C.exp.as_index_var();
+	    int index = regs.access(R).C.exp.as_index_var();
 
-	    int R2 = access(R).C.lookup_in_env( index );
+	    int R2 = regs.access(R).C.lookup_in_env( index );
 
 	    int R3 = incremental_evaluate_unchangeable( R2 );
 
@@ -584,20 +595,20 @@ int reg_heap::incremental_evaluate_unchangeable_(int R)
 	}
 
 	// Check for WHNF *OR* heap variables
-	else if (is_WHNF(access(R).C.exp))
-	    access(R).type = reg::type_t::constant;
+	else if (is_WHNF(regs.access(R).C.exp))
+	    regs.access(R).type = reg::type_t::constant;
 
 #ifndef NDEBUG
-	else if (access(R).C.exp.head().is_a<Trim>())
+	else if (regs.access(R).C.exp.head().is_a<Trim>())
 	    std::abort();
-	else if (access(R).C.exp.type() == parameter_type)
+	else if (regs.access(R).C.exp.type() == parameter_type)
 	    std::abort();
 #endif
 
 	// 3. Reduction: Operation (includes @, case, +, etc.)
 	else
 	{
-	    auto O = access(R).C.exp.head().assert_is_a<Operation>()->op;
+	    auto O = regs.access(R).C.exp.head().assert_is_a<Operation>()->op;
 
 	    // Although the reg itself is not a modifiable, it will stay changeable if it ever computes a changeable value.
 	    // Therefore, we cannot do "assert(not result_for_reg(t,R).changeable);" here.
@@ -605,7 +616,7 @@ int reg_heap::incremental_evaluate_unchangeable_(int R)
 #ifdef DEBUG_MACHINE
 	    string SS = "";
 	    SS = compact_graph_expression(*this, R, get_identifiers()).print();
-	    string SSS = untranslate_vars(deindexify(trim_unnormalize(access(R).C)),  
+	    string SSS = untranslate_vars(deindexify(trim_unnormalize(regs.access(R).C)),  
 					  get_identifiers()).print();
 	    if (log_verbose >= 3)
 		dot_graph_for_token(*this, root_token);
@@ -621,12 +632,12 @@ int reg_heap::incremental_evaluate_unchangeable_(int R)
 	    }
 	    catch (no_context&)
 	    {
-		access(R).type = reg::type_t::changeable;
+		regs.access(R).type = reg::type_t::changeable;
 		return R;
 	    }
 	    catch (error_exception& e)
 	    {
-		throw e;
+		throw;
 	    }
 	    catch (myexception& e)
 	    {
