@@ -198,7 +198,7 @@ namespace MCMC {
 	    o<<" move = "<<name<<"\n";
 	    o<<"   submove = "<<moves[order[i]]->name<<"\n";
 	    e.prepend(o.str());
-	    throw e;
+	    throw;
 	}
     }
 
@@ -317,7 +317,7 @@ namespace MCMC {
 	    std::ostringstream o;
 	    o<<" [single] move = "<<name<<"\n";
 	    e.prepend(o.str());
-	    throw e;
+	    throw;
 	}
     }
 
@@ -344,7 +344,7 @@ namespace MCMC {
 	    std::ostringstream o;
 	    o<<" [MH] move = "<<name<<"  (during proposal)\n";
 	    e.prepend(o.str());
-	    throw e;
+	    throw;
 	}
     
 
@@ -420,14 +420,12 @@ namespace MCMC {
 	    std::cerr<<endl<<endl;
 	}
 
-	double transformed_v1 = transform(v1);
-	double transformed_v2 = slice_sample(transformed_v1,slice_levels,W,100);
-	double v2 = inverse(transformed_v2);
+	double v2 = slice_sample(v1,slice_levels,W,100);
 
 	if (n_learning_iterations-- > 0) 
 	{
 	    n_tries++;
-	    total_movement += std::abs(transformed_v2 - transformed_v1);
+	    total_movement += std::abs(v2 - v1);
       
 	    double W_predicted = 4.0*total_movement/n_tries;
 	    if (n_tries > 3)
@@ -437,8 +435,9 @@ namespace MCMC {
 	if (log_verbose >= 3)
 	{
 	    show_parameters(std::cerr,P);
-	    std::cerr<<P.probability()<<" = "<<P.likelihood()<<" + "<<P.prior();
+	    std::cerr<<P.probability()<<" = "<<P.likelihood()<<" + "<<P.prior()<<std::endl;
 	    std::cerr<<endl<<endl;
+	    std::cerr<<" [Slice] window size = "<<W<<std::endl;
 	}
 
 	return v2;
@@ -460,8 +459,6 @@ namespace MCMC {
     Slice_Move::Slice_Move(const string& s)
 	:Move(s),
 	 W(1),
-	 transform(slice_sampling::identity),
-	 inverse(slice_sampling::identity),
 	 n_learning_iterations(0),
 	 n_tries(0),
 	 total_movement(0)
@@ -470,8 +467,6 @@ namespace MCMC {
     Slice_Move::Slice_Move(const string& s, const string& v)
 	:Move(s,v),
 	 W(1),
-	 transform(slice_sampling::identity),
-	 inverse(slice_sampling::identity),
 	 n_learning_iterations(0),
 	 n_tries(0),
 	 total_movement(0)
@@ -480,8 +475,6 @@ namespace MCMC {
     Slice_Move::Slice_Move(const string& s, const string& v, double W_)
 	:Move(s,v),
 	 W(W_),
-	 transform(slice_sampling::identity),
-	 inverse(slice_sampling::identity),
 	 n_learning_iterations(0),
 	 n_tries(0),
 	 total_movement(0)
@@ -490,20 +483,6 @@ namespace MCMC {
     Slice_Move::Slice_Move(const string& s, double W_)
 	:Move(s),
 	 W(W_),
-	 transform(slice_sampling::identity),
-	 inverse(slice_sampling::identity),
-	 n_learning_iterations(0),
-	 n_tries(0),
-	 total_movement(0)
-    {}
-
-    Slice_Move::Slice_Move(const string& s, const string& v, double W_,
-			   double(*f1)(double),
-			   double(*f2)(double))
-	:Move(s,v),
-	 W(W_),
-	 transform(f1),
-	 inverse(f2),
 	 n_learning_iterations(0),
 	 n_tries(0),
 	 total_movement(0)
@@ -515,15 +494,16 @@ namespace MCMC {
 
 	double v1 = P->get_modifiable_value(m_index).as_double();
 
-	modifiable_slice_function logp(*P, m_index, bounds, transform, inverse);
+	modifiable_slice_function logp(*P, bounds, m_index);
 
 	double v2 = sample(*P,logp,v1);
 
-	//---------- Record Statistics --------------//
+        //---------- Record Statistics --------------//
 	Result result(2);
 	result.totals[0] = std::abs(v2-v1);
 	result.totals[1] = logp.count;
     
+	if (log_verbose >= 3) std::cerr<<" [modifiable slice] posterior evaluated "<<logp.count<<" times."<<std::endl;
 	Stats.inc(name,result);
     }
 
@@ -537,20 +517,6 @@ namespace MCMC {
 	:Slice_Move(s,v,W_), m_index(m), bounds(b)
     {}
 
-    Modifiable_Slice_Move::Modifiable_Slice_Move(const string& s,int m,
-						 const Bounds<double>& b, double W_,
-						 double(*f1)(double),
-						 double(*f2)(double))
-	:Slice_Move(s,"",W_,f1,f2), m_index(m), bounds(b)
-    {}
-
-    Modifiable_Slice_Move::Modifiable_Slice_Move(const string& s, const string& v,int m,
-						 const Bounds<double>& b, double W_,
-						 double(*f1)(double),
-						 double(*f2)(double))
-	:Slice_Move(s,v,W_,f1,f2), m_index(m), bounds(b)
-    {}
-
     void Integer_Modifiable_Slice_Move::iterate(owned_ptr<Model>& P,MoveStats& Stats,int)
     {
 	if (log_verbose >= 3) clog<<" [integer modifiable slice] move = "<<m_index<<endl;
@@ -558,7 +524,7 @@ namespace MCMC {
 	int v1 = P->get_modifiable_value(m_index).as_int();
 	double x1 = double(v1)+uniform();
 
-	integer_modifiable_slice_function logp(*P, m_index, bounds, transform, inverse);
+	integer_modifiable_slice_function logp(*P, bounds, m_index);
 
 	double x2 = sample(*P,logp,x1);
 
@@ -567,6 +533,8 @@ namespace MCMC {
 	result.totals[0] = std::abs(x2-x1);
 	result.totals[1] = logp.count;
     
+	if (log_verbose >= 3) std::cerr<<"   - Posterior evaluated "<<logp.count<<" times."<<std::endl;
+
 	Stats.inc(name,result);
     }
 
@@ -580,20 +548,6 @@ namespace MCMC {
 	:Slice_Move(s,v,W_), m_index(m), bounds(b)
     {}
 
-    Integer_Modifiable_Slice_Move::Integer_Modifiable_Slice_Move(const string& s,int m,
-								 const Bounds<int>& b, double W_,
-								 double(*f1)(double),
-								 double(*f2)(double))
-	:Slice_Move(s,"",W_,f1,f2), m_index(m), bounds(b)
-    {}
-
-    Integer_Modifiable_Slice_Move::Integer_Modifiable_Slice_Move(const string& s, const string& v,int m,
-								 const Bounds<int>& b, double W_,
-								 double(*f1)(double),
-								 double(*f2)(double))
-	:Slice_Move(s,v,W_,f1,f2), m_index(m), bounds(b)
-    {}
-
     void Dirichlet_Slice_Move::iterate(owned_ptr<Model>& P,MoveStats& Stats,int)
     {
 	if (log_verbose >= 3) clog<<" [dirichlet slice] move"<<endl;
@@ -834,7 +788,7 @@ namespace MCMC {
 	    std::ostringstream o;
 	    o<<" [single] move = "<<name<<"\n";
 	    e.prepend(o.str());
-	    throw e;
+	    throw;
 	}
     }
     
@@ -1112,7 +1066,7 @@ namespace MCMC {
 	    if (subsample <= 0) subsample = 2*int(log(t.n_leaves()))+1;
 
 	    if (alignment_burnin_iterations > 0)
-		PP->set_parameter_value(PP->find_parameter("*IModels.training"), new constructor("Prelude.True",0));
+		PP->set_parameter_value(PP->find_parameter("*IModels.training"), bool_true);
 	}
 
 	//---------------- Run the MCMC chain -------------------//
@@ -1122,7 +1076,7 @@ namespace MCMC {
 	    {
 		// Free temporarily fixed parameters at iteration 5
 		if (iterations == alignment_burnin_iterations)
-		    PP->set_parameter_value(PP->find_parameter("*IModels.training"), new constructor("Prelude.False",0));
+		    PP->set_parameter_value(PP->find_parameter("*IModels.training"), bool_false);
 
 		// Change the temperature according to the pattern suggested
 		if (iterations < PP->PC->beta_series.size())
@@ -1132,6 +1086,11 @@ namespace MCMC {
 		if (iterations == 5)
 		    start_learning(100);
 	    }
+	    else
+	    {
+		if (iterations == 0)
+		    start_learning(100);
+	    }
 
 	    // Stop learning step sizes at iteration 500
 	    if (iterations == 500)
