@@ -9,9 +9,11 @@
 #include "models/rules.H"
 #include "startup/paths.H"
 #include "startup/help.hh"
+#include "util/text.H"
 #include "computation/expression/lambda.H"
 #include <boost/optional/optional_io.hpp>
 #include <boost/filesystem/operations.hpp>
+#include "sequence/sequence-format.H"
 
 namespace fs = boost::filesystem;
 
@@ -27,6 +29,7 @@ using std::endl;
 using std::ostream;
 using std::string;
 using std::vector;
+using std::pair;
 using boost::optional;
 
 /// Replace negative or zero branch lengths with saner values.
@@ -146,12 +149,44 @@ json optional_to_json(const boost::optional<T>& o)
 }
 
 
+pair<string,string> split_on_last(char sep, const string& s)
+{
+    string s1 = s;
+    string s2;
+    auto pos = s1.rfind(sep);
+    if (pos != string::npos)
+    {
+	s2 = s.substr(pos+1);
+	s1 = s1.substr(0,pos);
+    }
+    return {s1,s2};
+}
+
+vector<pair<string,string>> split_on_last(char sep, const vector<string>& v1)
+{
+    vector<pair<string,string>> v2;
+    for(auto& s: v1)
+	v2.push_back(split_on_last(sep, s));
+    return v2;
+}
+
 // FIXME - maybe we should try to make a single giant model so that we get S1/parameter
 //           exactly when this occurs for the logged parameter names, themselves.
 //
 //           This would also help it we have some variables that are outside of models, because
 //             they are shared between them.
 
+string tag(string s, int i)
+{
+    return s+convertToString(i+1);
+}
+
+string show_model(const model_t& m)
+{
+    string s = indent_and_wrap(0,12,1000,m.show_main(false));
+    return s.substr(0,2) + bold(s.substr(2));
+}
+
 json log_summary(ostream& out_cache, ostream& out_screen,ostream& out_both,
 		 const vector<model_t>& IModels, const vector<model_t>& SModels,
 		 const vector<model_t>& ScaleModels,
@@ -175,40 +210,46 @@ json log_summary(ostream& out_cache, ostream& out_screen,ostream& out_both,
     }
 
     //-------- Log some stuff -----------//
-    vector<string> filenames = args["align"].as<vector<string> >();
+    auto filename_ranges = args["align"].as<vector<string> >();
+    auto alignment_files = split_on_last(':',filename_ranges);
 
     for(int i=0;i<P.n_data_partitions();i++)
     {
 	json partition;
 
+	out_screen<<"Partition "<<magenta(tag("P",i))<<":\n";
 	// 1. filename 
-	out_cache<<"data"<<i+1<<" = "<<filenames[i]<<endl;
-	out_screen<<"#"<<i+1<<": file = "<<filenames[i]<<endl;
-	partition["filename"] = filenames[i];
+	out_cache<<"data"<<i+1<<" = "<<filename_ranges[i]<<endl;
+	string file = bold(alignment_files[i].first);
+	if (alignment_files[i].second.size())
+	    file += ":"+alignment_files[i].second;
+	out_screen<<"    file = "<<bold(file)<<endl;
+	partition["filename"] = alignment_files[i].first;
+	partition["range"] = alignment_files[i].second;
 
 	// 2. alphabet
 	string a_name = P[i].get_alphabet().name;
-	out_screen<<"#"<<i+1 <<": alphabet = "<<a_name<<"\n";
+	out_screen<<"    alphabet = "<<bold(a_name)<<"\n";
 	out_cache<<"alphabet"<<i+1<<" = "<<a_name<<endl;
 	partition["alphabet"] = a_name;
 
 	// 3. substitution model
 	auto s_index = P.smodel_index_for_partition(i);
-	out_screen<<"#"<<i+1<<": subst "<<indent_and_wrap(0,12,1000,SModels[*s_index].show_main(false))<<" (S"<<*s_index+1<<")\n";
+	out_screen<<"    subst "<<show_model(SModels[*s_index])<<" ("<<bold_blue(tag("S",*s_index))<<")\n";
 	out_cache<<"smodel-index"<<i+1<<" = "<<P.smodel_index_for_partition(i)<<endl;
 	partition["smodel"] = optional_to_json( P.smodel_index_for_partition(i) );
 
 	// 4. indel model
 	if (auto i_index = P.imodel_index_for_partition(i))
-	    out_screen<<"#"<<i+1<<": indel "<<indent_and_wrap(0,12,1000,IModels[*i_index].show_main(false))<<" (I"<<*i_index+1<<")\n";
+	    out_screen<<"    indel "<<show_model(IModels[*i_index])<<" ("<<bold_red(tag("I",*i_index))<<")\n";
 	else
-	    out_screen<<"#"<<i+1<<": indel = none\n";
+	    out_screen<<"    indel = "<<bold("none")<<"\n";
 	out_cache<<"imodel-index"<<i+1<<" = "<<P.imodel_index_for_partition(i)<<endl;
 	partition["imodel"] = optional_to_json( P.imodel_index_for_partition(i) );
 
 	// 5. scale model
 	auto scale_index = P.scale_index_for_partition(i);
-	out_screen<<"#"<<i+1<<": scale "<<indent_and_wrap(0,12,1000,ScaleModels[*scale_index].show_main(false))<<" (Scale"<<*scale_index+1<<")\n";
+	out_screen<<"    scale "<<show_model(ScaleModels[*scale_index])<<" ("<<green(tag("Scale",*scale_index))<<")\n";
 	out_cache<<"scale-index"<<i+1<<" = "<<P.scale_index_for_partition(i)<<endl;
 	partition["scale"] = optional_to_json( P.scale_index_for_partition(i) );
 
@@ -225,7 +266,7 @@ json log_summary(ostream& out_cache, ostream& out_screen,ostream& out_both,
 	smodels.push_back(SModels[i].pretty_model());
 	string e = SModels[i].show_extracted();
 	if (e.size())
-	    out_screen<<"Substitution model (S"<<i+1<<") -- priors:"<<e<<"\n\n";
+	    out_screen<<"Substitution model "<<bold_blue(tag("S",i))<<" priors:"<<e<<"\n\n";
     }
 
     json imodels = json::array();
@@ -235,7 +276,7 @@ json log_summary(ostream& out_cache, ostream& out_screen,ostream& out_both,
 	imodels.push_back(IModels[i].pretty_model());
 	string e = IModels[i].show_extracted();
 	if (e.size())
-	    out_screen<<"Insertion/deletion model (I"<<i+1<<") -- priors:"<<e<<"\n\n";
+	    out_screen<<"Insertion/deletion model "<<bold_red(tag("I",i))<<" priors:"<<e<<"\n\n";
     }
 
     json scales = json::array();
@@ -245,7 +286,7 @@ json log_summary(ostream& out_cache, ostream& out_screen,ostream& out_both,
 	scales.push_back(ScaleModels[i].pretty_model());
 	string e = ScaleModels[i].show_extracted();
 	if (e.size())
-	    out_screen<<"Scale model (Scale"<<i+1<<") -- priors:"<<e<<"\n\n";
+	    out_screen<<"Scale model "<<green(tag("Scale",i))<<" priors:"<<e<<"\n\n";
     }
 
     info["partitions"] = partitions;
@@ -270,7 +311,7 @@ void check_alignment_names(const alignment& A)
     }
 }
 
-void check_alignment_values(const alignment& A,const string& filename)
+void check_alignment_values(const alignment& A,const pair<string,string>& filename_range)
 {
     const alphabet& a = A.get_alphabet();
 
@@ -280,7 +321,12 @@ void check_alignment_values(const alignment& A,const string& filename)
 
 	for(int j=0;j<A.length();j++) 
 	    if (A.unknown(j,i))
-		throw myexception()<<"Alignment file '"<<filename<<"' has a '"<<a.unknown_letter<<"' in sequence '"<<name<<"'.\n (Please replace with gap character '"<<a.gap_letter<<"' or wildcard '"<<a.wildcard<<"'.)";
+	    {
+		string file = filename_range.first;
+		if (filename_range.second.size())
+		    file = file + ":" + filename_range.second;
+		throw myexception()<<"Alignment file '"<<file<<"' has a '"<<a.unknown_letter<<"' in sequence '"<<name<<"'.\n (Please replace with gap character '"<<a.gap_letter<<"' or wildcard '"<<a.wildcard<<"'.)";
+	    }
     }
 }
 
@@ -401,22 +447,39 @@ alignment unalign_A(const alignment& A)
     return get_alignment(unaligned_matrix(L),A);
 }
 
+bool needs_constant_alignment(const alphabet& a)
+{
+    if (dynamic_cast<const Doublets*>(&a)) return true;
+    if (dynamic_cast<const Numeric*>(&a))  return true;
+
+    return false;
+}
+
+bool can_share_imodel(const alphabet& a1, const alphabet& a2)
+{
+    if (needs_constant_alignment(a1) != needs_constant_alignment(a2)) return false;
+
+    if (a1.width() != a2.width()) return false;
+
+    if (dynamic_cast<const Nucleotides*>(&a1)) return bool(dynamic_cast<const Nucleotides*>(&a2));
+
+    if (dynamic_cast<const Doublets*>(&a1))    return bool(dynamic_cast<const Doublets*>(&a2));
+
+    if (dynamic_cast<const Triplets*>(&a1))    return bool(dynamic_cast<const Triplets*>(&a2));
+
+    if (dynamic_cast<const AminoAcids*>(&a1))  return bool(dynamic_cast<const AminoAcids*>(&a2));
+
+    return a1.name == a2.name;
+}
 
 owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const std::shared_ptr<module_loader>& L,
 				      ostream& out_cache, ostream& out_screen, ostream& out_both, json& info,
 				      int proc_id)
 {
     //------ Determine number of partitions ------//
-    vector<string> filenames = args["align"].as<vector<string> >();
-    const int n_partitions = filenames.size();
+    auto alignment_files = split_on_last(':', args["align"].as<vector<string> >() );
 
-    //-------------Choose an indel model--------------//
-    auto imodel_names_mapping = get_mapping(args, "imodel", n_partitions);
-    auto& imodel_mapping = imodel_names_mapping.item_for_partition;
-
-    for(int i=0;i<imodel_names_mapping.n_unique_items();i++)
-	if (imodel_names_mapping.unique(i) == "")
-	    imodel_names_mapping.unique(i) = "rs07";
+    const int n_partitions = alignment_files.size();
 
     //------------- Get smodel names -------------------
     auto smodel_names_mapping = get_mapping(args, "smodel", n_partitions);
@@ -430,7 +493,7 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 	    full_smodels[i] = get_model(R, "MultiMixtureModel[a]",smodel_names_mapping.unique(i));
 
     //------------- Get alphabet names -------------------
-    shared_items<string> alphabet_names_mapping = get_mapping(args, "alphabet", filenames.size());
+    shared_items<string> alphabet_names_mapping = get_mapping(args, "alphabet", alignment_files.size());
     vector<string> alphabet_names;
     for(int i=0;i<alphabet_names_mapping.n_partitions();i++)
 	alphabet_names.push_back(alphabet_names_mapping[i]);
@@ -474,6 +537,11 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 		if (a != "Triplets")
 		    throw myexception()<<"Partition "<<a_specified[0]+1<<" has specified alphabet '"<<a<<"' but the substitution model requires a triplet alphabet!";
 	    }
+	    else if (alphabet_type.get_value<string>() == "Doublets")
+	    {
+		if (a != "Doublets")
+		    throw myexception()<<"Partition "<<a_specified[0]+1<<" has specified alphabet '"<<a<<"' but the substitution model requires a doublet alphabet!";
+	    }
 	    else if (alphabet_type.get_value<string>() == "AA")
 	    {
 		if (a != "AA" and a != "Amino-Acids" and a != "Amino-AcidsWithStop")
@@ -490,6 +558,11 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 	    for(int j: smodel_names_mapping.partitions_for_item[i])
 		alphabet_names[j] = "Triplets";
 	}
+	else if (alphabet_type.get_value<string>() == "Doublets")
+	{
+	    for(int j: smodel_names_mapping.partitions_for_item[i])
+		alphabet_names[j] = "Doublets";
+	}
 	else
 	{
 	    bool triplet_constraint = false;
@@ -500,8 +573,17 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 	    if (triplet_constraint)
 		for(int j: smodel_names_mapping.partitions_for_item[i])
 		    alphabet_names[j] = "Triplets";
+
+	    bool doublet_constraint = false;
+	    for(auto& constraint: constraints)
+		if (constraint.get_value<string>() == "Doublets" and constraint.begin()->second.get_value<string>() == "a")
+		    doublet_constraint = true;
+
+	    if (doublet_constraint)
+		for(int j: smodel_names_mapping.partitions_for_item[i])
+		    alphabet_names[j] = "Doublets";
 	}
-//      Use the auto-detected alphabet right now -- it leaves to better error messages.
+//      Use the auto-detected alphabet right now -- it leads to better error messages.
 //	else if (alphabet_type.get_value<string>() == "AA")
 //	{
 //	    for(int j: smodel_names_mapping.partitions_for_item[i])
@@ -510,26 +592,20 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
     }
 
     //----------- Load alignments  ---------//
-    vector<alignment> A(filenames.size());
+    vector<alignment> A(alignment_files.size());
 
-    // 3. -- load alignments for SPECIFIED alphabets
-    for(int i=0;i<filenames.size();i++)
-	if (alphabet_names[i].size())
-	    // FIXME - if we are loading codons, FIRST determine the UNDERLYING alphabet, then
-	    // But actually we DO this inside the routine, right?
-	    A[i] = load_alignment(filenames[i], alphabet_names[i]);
-
-    // 4. -- load alignments for UNSPECIFIED alphabets && set alphabet names.
-    for(int i=0;i<filenames.size();i++)
+    // 3. -- load alignments for SPECIFIED and UNSPECIFIED alphabets
+    for(int i=0;i<alignment_files.size();i++)
     {
-	if (not A[i].has_alphabet())
-	    A[i] = load_alignment(filenames[i]);
-	alphabet_names[i] = A[i].get_alphabet().name;
+	A[i] = load_alignment_with_range(alignment_files[i].first, alignment_files[i].second, alphabet_names[i]);
+
+	if (alphabet_names[i].empty())
+	    alphabet_names[i] = A[i].get_alphabet().name;
     }
 
     for(int i=0;i<A.size();i++) {
 	check_alignment_names(A[i]);
-	check_alignment_values(A[i],filenames[i]);
+	check_alignment_values(A[i],alignment_files[i]);
     }
 
     //----------- Load tree and link to alignments ---------//
@@ -556,6 +632,52 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
     sanitize_branch_lengths(T);
 
     //--------- Set up indel model --------//
+    auto imodel_names_mapping = get_mapping(args, "imodel", n_partitions);
+    auto& imodel_mapping = imodel_names_mapping.item_for_partition;
+
+    //-- Check that we're not estimating the alignment for things that aren't sequences --//
+    for(int i = imodel_names_mapping.n_unique_items() - 1; i >= 0; i--)
+    {
+	auto& value = imodel_names_mapping.unique(i);
+
+	// Maybe complain based on needs_constant_alignment(a)
+
+        // Check that all partitions with the same indel model have compatible alphabets
+
+	bool need_constant_a = false;
+	optional<int> j0;
+	for(int j: imodel_names_mapping.partitions_for_item[i])
+	{
+	    auto& a = A[j].get_alphabet();
+
+	    if (not j0) j0 = j;
+
+	    if (value.empty() and needs_constant_alignment(A[*j0].get_alphabet()) != needs_constant_alignment(a))
+		throw myexception()<<"Can't guess shared indel model for incompatible partitions "<<*j0+1<<" ("<<A[*j0].get_alphabet().name<<") and "<<j+1<<" ("<<a.name<<").\n";
+
+	    if (not can_share_imodel(A[*j0].get_alphabet(), a))
+		std::cerr<<"Warning! Linking indel model '"<<value<<"' for partition "<<*j0+1<<" ("<<A[*j0].get_alphabet().name<<") and partition "<<j+1<<" ("<<a.name<<") -- is this right?\n";
+
+	    if (needs_constant_alignment(a))
+	    {
+		if (not value.empty())
+		    throw myexception()<<"Data of type '"<<a.name<<"' requires a constant alignment in partition "<<j+1<<", but got indel model '"<<value<<"'";
+		need_constant_a = true;
+	    }
+	}
+
+	// Choose default indel model based on needs_constant_alignment(a)
+	if (value == "")
+	{
+	    if (need_constant_a)
+	    {
+		imodel_names_mapping.remove_nth_item(i);
+		break;
+	    }
+	    else
+		value = "rs07";
+	}
+    }
     auto full_imodels = get_imodels(R, imodel_names_mapping, T);
 
     // 5. ----- Check that all smodel-linked partitions end up with the same alphabet. -----
@@ -608,6 +730,9 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 	if (alphabet_type.get_value<string>() == "Triplets" and not dynamic_cast<const Triplets*>(&a))
 	    throw myexception()<<"Substitution model S"<<i+1<<" requires a triplet alphabet, but sequences are '"<<a.name<<"'";;
 
+	if (alphabet_type.get_value<string>() == "Doublets" and not dynamic_cast<const Doublets*>(&a))
+	    throw myexception()<<"Substitution model S"<<i+1<<" requires a doublet alphabet, but sequences are '"<<a.name<<"'";;
+
 	if (alphabet_type.get_value<string>() == "AA" and not dynamic_cast<const AminoAcids*>(&a))
 	    throw myexception()<<"Substitution model S"<<i+1<<" requires an amino-acid alphabet, but sequences are '"<<a.name<<"'";;
     }
@@ -716,18 +841,22 @@ owned_ptr<Model> create_A_and_T_model(const Rules& R, variables_map& args, const
 
 void write_initial_alignments(variables_map& args, int proc_id, const string& dir_name)
 {
-    vector<string> filenames = args["align"].as<vector<string> >();
+    auto alignment_files = split_on_last(':', args["align"].as<vector<string> >() );
 
     fs::path dir(dir_name);
 
     string base = "C" + convertToString(proc_id+1);
 
-    int i=0;
-    for(auto& filename: filenames)
+    int i=1;
+    for(auto& alignment_file: alignment_files)
     {
-	auto source = fs::path(filename);
-	auto target = fs::path(base+".P"+convertToString(i+1)+".initial.fasta");
-	fs::copy_file(source,dir/target);
+	auto sequences = load_sequences_with_range(alignment_file.first, alignment_file.second);
+
+	auto target = fs::path(base+".P"+convertToString(i)+".initial.fasta");
+	target = dir/target;
+	checked_ofstream out(target.string(), false);
+	sequence_format::write_fasta(out, sequences);
+
 	i++;
     }
 }
